#ifndef CMS_VARIABLE_CXX
#define CMS_VARIABLE_CXX

//----------------------------------
// Includes and forward declarations
//----------------------------------
#include "cyg/io/cmsvar.h"
#include <cyg/kal/kdefs.h>
#include <string.h>

//---------------------------------------
// Define the Variable command specifiers
//---------------------------------------
#define CS_READ  0x80
#define CS_WRITE 0x00

//---------------------------------
// Define the Variable result codes
//---------------------------------
#define RC_FAILURE 0x80
#define RC_SUCCESS 0x00

//---------------------
// Class implementation
//---------------------
// CMS_Variable
// Constructor
CMS_Variable::CMS_Variable(CMS_Variable::CMS_VariableAccessType a_accesstype,
                           CMS_Variable::CMS_VariableClass a_class,
                           unsigned char a_length)
    : m_accesstype(a_accesstype),
      m_class(a_class),
      m_length(a_length)
{
}

// Provide an (empty) virtual destructor
CMS_Variable::~CMS_Variable()
{
}

// CMS_ServerVariable 
// Constructor
CMS_ServerVariable::CMS_ServerVariable(CMS_Variable::CMS_VariableAccessType a_accesstype,
                                       CMS_Variable::CMS_VariableClass a_class,
                                       unsigned int a_inhibittime,
                                       unsigned char a_priority,
                                       const char * a_name,
                                       unsigned char a_length,
                                       bool a_numeric)
    : CMS_Variable(a_accesstype, a_class, a_length),
      CMS_ServerObject(a_inhibittime, a_priority, a_name, a_numeric)
{
}

// Provide an (empty) destructor
CMS_ServerVariable::~CMS_ServerVariable()
{
}

// update
void CMS_ServerVariable::update(const unsigned char * a_data)
{
  if(m_accesstype == Read_Only)
  {
    updateData(a_data);
  }
}

// dataIndication
void CMS_ServerVariable::dataIndication(unsigned char a_data[8],
                                        unsigned char a_length,
                                        CommunicationObject * /*a_cob*/)
{
  if(m_accesstype == Read_Only)
  {
    readIndication(a_data, a_data[0] & 0x7f);
    return;
  }

  if(m_accesstype == Write_Only)
  {
    writeIndication(a_data, a_data[0] & 0x7f);
    return;
  }

  if(m_accesstype == Read_Write)
  {
    if(a_data[0] & CS_READ)
      readIndication(a_data, a_data[0] & 0x7f);
    else
      writeIndication(a_data, a_data[0] & 0x7f);
    return;
  }
}

// CMS_ClientVariable
// Constructor
CMS_ClientVariable::CMS_ClientVariable(CMS_Variable::CMS_VariableAccessType a_accesstype,
                                       CMS_Variable::CMS_VariableClass a_class,
                                       unsigned int a_inhibittime,
                                       unsigned char a_priority,
                                       const char * a_name,
                                       unsigned char a_length,
                                       bool a_numeric)
    : CMS_Variable(a_accesstype, a_class, a_length),
      CMS_ClientObject(a_inhibittime, a_priority, a_name, a_numeric)
{
  memset(&m_data[0], 0, 8);
}
      
// Provide an (empty) virtual destructor
CMS_ClientVariable::~CMS_ClientVariable()
{
}

// read
CMS_Error CMS_ClientVariable::read(unsigned char * a_rdata,
                                   unsigned char a_mux, 
                                   unsigned int a_timeout)
{
  CMS_Error result;

  if(m_accesstype == Write_Only)
    return -CMS_ACCESSERR;

  KSingleLock sl1(&m_cond);
  readRequest(a_mux);
  
  if(!sl1.lock(a_timeout))
    return -CMS_TIMEOUT;

  sl1.unlock();

  KSingleLock sl2(&m_mux);
  sl2.lock();
  result = readConfirmation(a_rdata, a_mux);
  sl2.unlock();

  return result;
}

// write
CMS_Error CMS_ClientVariable::write(unsigned char * a_wdata,
                                    unsigned char a_mux,
                                    unsigned int a_timeout)
{
  CMS_Error result;
  if(m_accesstype == Read_Only)
    return -CMS_ACCESSERR;

  if(m_accesstype == Write_Only)
  {
    writeRequest(a_wdata, a_mux);
    return -CMS_NOERR;
  }

  KSingleLock sl1(&m_cond);
  writeRequest(a_wdata, a_mux);

  if(!sl1.lock(a_timeout))
    return -CMS_TIMEOUT;

  sl1.unlock();

  KSingleLock sl2(&m_mux);
  sl2.lock();
  result = writeConfirmation(a_wdata, a_mux);
  sl2.unlock();

  return result;
}

// dataIndication
void CMS_ClientVariable::dataIndication(unsigned char a_data[8],
                                        unsigned char a_length,
                                        CommunicationObject * /*a_cob*/)
{
  memcpy(&m_data[0], a_data, a_length);
  m_cond.signal();
}

// CMS_BasicServerVariable
// Constructor
CMS_BasicServerVariable::CMS_BasicServerVariable(CMS_Variable::CMS_VariableAccessType a_accesstype,
                                                 unsigned int a_inhibittime,
                                                 unsigned char a_priority,
                                                 const char * a_name,
                                                 unsigned char a_length,
                                                 CAN_msgid a_txcobid,
                                                 CAN_msgid a_rxcobid,
                                                 can_io_handle_t a_iohandle,
                                                 bool a_numeric)
    : CMS_ServerVariable(a_accesstype, 
                         CMS_Variable::Basic, 
                         a_inhibittime, 
                         a_priority, 
                         a_name, 
                         a_length, 
                         a_numeric)
{
  switch(a_accesstype)
  {
    case Read_Only:
      m_txcob = new RTR_ServerCOB(a_txcobid, a_iohandle, this);
      _ASSERT(a_length < 9, "Invalid datalength specified");
      break;
    case Write_Only:
      m_rxcob = new RxCOB(a_rxcobid, a_iohandle, this);
      _ASSERT(a_length < 9, "Invalid datalength specified");
      break;
    case Read_Write:
      m_rxcob = new RxCOB(a_rxcobid, a_iohandle, this);
      m_txcob = new TxCOB(a_txcobid, a_iohandle, this);
      break;
      _ASSERT(a_length < 8, "Invalid datalength specified");
    default:
      break;
  }
}

// Provide an (empty) virtual destructor
CMS_BasicServerVariable::~CMS_BasicServerVariable()
{
}

// readIndication
void CMS_BasicServerVariable::readIndication(unsigned char * a_buffer, unsigned char/* a_mux*/)
{
  a_buffer[0] = 0;

  if(!readResponse(&a_buffer[1]))
    a_buffer[0] |= RC_FAILURE;

  ((TxCOB *)m_txcob)->transmit(a_buffer, m_length + 1);
}

// writeIndication
void CMS_BasicServerVariable::writeIndication(unsigned char * a_buffer, unsigned char /*a_mux*/)
{
  switch(m_accesstype)
  {
    case Write_Only:
      writeResponse(a_buffer);
      break;
    case Read_Write:
      {
        a_buffer[0] = 0;
        if(!writeResponse(&a_buffer[1]))
          a_buffer[0] |= RC_FAILURE;

        ((TxCOB *)m_txcob)->transmit(a_buffer, m_length + 1);
      }
      break;
    case Read_Only:
    default:
      break;
  }
}

// updateData
void CMS_BasicServerVariable::updateData(const unsigned char * a_data)
{
  unsigned char data[8];
  encode(&data[0], a_data, m_length);
  ((RTR_ServerCOB *)m_txcob)->update(data, m_length);
}

// CMS_BasicClientVariable
// Constructor
CMS_BasicClientVariable::CMS_BasicClientVariable(CMS_Variable::CMS_VariableAccessType a_accesstype,
                                                 unsigned int a_inhibittime,
                                                 unsigned char a_priority,
                                                 const char * a_name,
                                                 unsigned char a_length,
                                                 CAN_msgid a_txcobid,
                                                 CAN_msgid a_rxcobid,
                                                 can_io_handle_t a_iohandle,
                                                 bool a_numeric)
    : CMS_ClientVariable(a_accesstype, 
                         CMS_Variable::Basic,
                         a_inhibittime, 
                         a_priority, 
                         a_name, 
                         a_length, 
                         a_numeric)
{
  switch(a_accesstype)
  {
    case Read_Only:
      m_rxcob = new RTR_ClientCOB(a_rxcobid, a_iohandle, this, a_inhibittime);
      _ASSERT(a_length < 9, "Invalid datalength specified");
      break;
    case Write_Only:
      m_txcob = new TxCOB(a_txcobid, a_iohandle, this, a_inhibittime);
      _ASSERT(a_length < 9, "Invalid datalength specified");
      break;
    case Read_Write:
      m_txcob = new TxCOB(a_txcobid, a_iohandle, this, a_inhibittime);
      m_rxcob = new RxCOB(a_rxcobid, a_iohandle, this);
      _ASSERT(a_length < 8, "Invalid datalength specified");
      break;
    default:
      break;
  }
}

// Provide an (empty) virtual destructor
CMS_BasicClientVariable::~CMS_BasicClientVariable()
{
}

// readRequest
void CMS_BasicClientVariable::readRequest(unsigned char /*a_mux*/)
{
  switch(m_accesstype)
  {
    case Read_Only:
      ((RTR_ClientCOB *)m_rxcob)->request(m_length);
      break;
    case Read_Write:
      {
      unsigned char request = 0;
      request |= CS_READ;
      ((TxCOB *)m_txcob)->transmit(&request, m_length + 1);
      }
      break;
    case Write_Only:
    default:
      break;
  }
}

// writeRequest
void CMS_BasicClientVariable::writeRequest(const unsigned char * a_wdata, unsigned char /*a_mux*/)
{
  unsigned char data[8];

  switch(m_accesstype)
  {
    case Write_Only:
      {
      encode(&data[0], a_wdata, m_length);
      ((TxCOB *)m_txcob)->transmit(&data[0], m_length);
      }
      break;
    case Read_Write:
      {
      encode(&data[1], a_wdata, m_length);
      data[0] = 0;
      data[0] |= CS_WRITE;
      ((TxCOB *)m_txcob)->transmit(&data[0], m_length + 1);
      }
      break;
    case Read_Only:
    default:
      break;
  }
}

// readConfirmation
CMS_Error CMS_BasicClientVariable::readConfirmation(unsigned char * a_response, unsigned char /*a_mux*/)
{
  CMS_Error result = -CMS_NOERR;

  switch(m_accesstype)
  {
    case Read_Only:
      decode(a_response, &m_data[0], m_length);
      break;
    case Read_Write:
      {
      if(m_data[0] & RC_FAILURE)
      {
        memcpy(a_response, &m_data[1], m_length);
        result = -CMS_APPLICATIONERR;
      }
      else
        decode(a_response, &m_data[1], m_length);
      }
      break;
    case Write_Only:
    default:
      break;
  }

  return result;
}

// writeConfirmation
CMS_Error CMS_BasicClientVariable::writeConfirmation(unsigned char * a_response, unsigned char /*a_mux*/)
{
  CMS_Error result = -CMS_NOERR;

  if(m_data[0] & RC_FAILURE)
  {
    memcpy(a_response, &m_data[1], m_length);
    result = -CMS_APPLICATIONERR;
  }
  else
    decode(a_response, &m_data[1], m_length);

  return result;
}

// CMS_MultiplexedServerVariable
// Constructor
CMS_MultiplexedServerVariable::CMS_MultiplexedServerVariable(CMS_Variable::CMS_VariableAccessType a_accesstype,
                                                             unsigned int a_inhibittime,
                                                             unsigned char a_priority,
                                                             const char * a_name,
                                                             unsigned char a_length,
                                                             CAN_msgid a_txcobid,
                                                             CAN_msgid a_rxcobid,
                                                             can_io_handle_t a_iohandle,
                                                             bool a_numeric)
    : CMS_ServerVariable(a_accesstype, 
                         CMS_Variable::Multiplexed,
                         a_inhibittime, 
                         a_priority, 
                         a_name, 
                         a_length, 
                         a_numeric)
{
  switch(a_accesstype)
  {
    case Read_Only:
    case Read_Write:
      m_txcob = new TxCOB(a_txcobid, a_iohandle, this);
      m_rxcob = new RxCOB(a_rxcobid, a_iohandle, this);
      _ASSERT(a_length < 8, "Invalid datalength specified");
      break;
    case Write_Only:
      m_rxcob = new RxCOB(a_rxcobid, a_iohandle, this);
      _ASSERT(a_length < 8, "Invalid datalength specified");
      break;
    default:
      break;
  }
}

// Provide an (empty) virtual destructor
CMS_MultiplexedServerVariable::~CMS_MultiplexedServerVariable()
{
}

// readIndication
void CMS_MultiplexedServerVariable::readIndication(unsigned char * a_buffer,
                                                   unsigned char a_mux)
{
  a_buffer[0] = a_mux;
  if(!readResponse(&a_buffer[1], a_mux))
    a_buffer[0] |= RC_FAILURE;

  ((TxCOB *)m_txcob)->transmit(a_buffer, m_length + 1);
}

// writeIndication
void CMS_MultiplexedServerVariable::writeIndication(unsigned char * a_buffer,
                                                    unsigned char a_mux)
{
  switch(m_accesstype)
  {
    case Write_Only:
      {
      a_buffer[0] = a_mux;
      writeResponse(&a_buffer[1], a_mux);
      }
      break;
    case Read_Write:
      {
      a_buffer[0] = a_mux;
      if(!writeResponse(&a_buffer[1], a_mux))
        a_buffer[0] |= RC_FAILURE;

      ((TxCOB *)m_txcob)->transmit(a_buffer, m_length + 1);
      }
      break;
    case Read_Only:
    default:
      break;
  }
}

// CMS_MultiplexedClientVariable
// Constructor
CMS_MultiplexedClientVariable::CMS_MultiplexedClientVariable(CMS_Variable::CMS_VariableAccessType a_accesstype,
                                                             unsigned int a_inhibittime,
                                                             unsigned char a_priority,
                                                             const char * a_name,
                                                             unsigned char a_length,
                                                             CAN_msgid a_txcobid,
                                                             CAN_msgid a_rxcobid,
                                                             can_io_handle_t a_iohandle,
                                                             bool a_numeric)
    : CMS_ClientVariable(a_accesstype, 
                         CMS_Variable::Multiplexed,
                         a_inhibittime, 
                         a_priority, 
                         a_name, 
                         a_length, 
                         a_numeric)
{
  switch(a_accesstype)
  {
    case Read_Only:
    case Read_Write:
      m_txcob = new TxCOB(a_txcobid, a_iohandle, this, a_inhibittime);
      m_rxcob = new RxCOB(a_rxcobid, a_iohandle, this);
      _ASSERT(a_length < 8, "Invalid datalength specified");
      break;
    case Write_Only:
      m_txcob = new TxCOB(a_txcobid, a_iohandle, this, a_inhibittime);
      _ASSERT(a_length < 8, "Invalid datalength specified");
      break;
    default:
      break;
  }
}

// Provide an (empty) virtual destructor
CMS_MultiplexedClientVariable::~CMS_MultiplexedClientVariable()
{
}

// readRequest
void CMS_MultiplexedClientVariable::readRequest(unsigned char a_mux)
{
  unsigned char request = (a_mux & 0x7f);
  if(m_accesstype == Read_Write)
  {
    request |= CS_READ;
    ((TxCOB *)m_txcob)->transmit(&request, m_length + 1);
  }
  else
    ((TxCOB *)m_txcob)->transmit(&request, 1);
}

// readConfirmation
CMS_Error CMS_MultiplexedClientVariable::readConfirmation(unsigned char * a_response, unsigned char /*a_mux*/)
{
  CMS_Error result = -CMS_NOERR;

  if(m_data[0] & RC_FAILURE)
  {
    memcpy(a_response, &m_data[1], m_length);
    result = -CMS_APPLICATIONERR;
  }
  else
    decode(a_response, &m_data[1], m_length);

  return result;
}

// writeRequest
void CMS_MultiplexedClientVariable::writeRequest(const unsigned char * a_wdata, unsigned char a_mux)
{
  unsigned char data[8];
  data[0] = (a_mux & 0x7f);

  if(m_accesstype == Read_Write)
    data[0] |= CS_WRITE;

  encode(&data[1], a_wdata, m_length);
  ((TxCOB *)m_txcob)->transmit(&data[0], m_length + 1);
}

// writeConfirmation
CMS_Error CMS_MultiplexedClientVariable::writeConfirmation(unsigned char * a_response, unsigned char a_mux)
{
  CMS_Error result = -CMS_NOERR;
  
  if(m_data[0] & RC_FAILURE)
  {
    memcpy(a_response, &m_data[1], m_length);
    result = -CMS_APPLICATIONERR;
  }
  else
    decode(a_response, &m_data[1], m_length);

  return result;
}

#endif // CMS_VARIABLE_CXX
