#ifndef COMOBJ_CXX
#define COMOBJ_CXX

//----------------------------------
// Includes and forward declarations
//----------------------------------
#include "cyg/io/comobj.h"
#include "cyg/io/canmsgdisp.h"
#include "cyg/io/calio.h"
#include <cyg/kal/kdefs.h>
#include <cyg/kal/ktimer.h>

//-----------------------------------
// Allocate memory for static members
//-----------------------------------
unsigned long CommunicationObject::m_tickcount = 0;
KCondition * CommunicationObject::m_tickcondition;

//--------------------------------------------------------------
// A special (not visible in the interface) Communication object
//--------------------------------------------------------------
class TimerCOB : public KTimerClient,
                 public CommunicationObject
{
  public:
                      TimerCOB(KCondition * a_condition)
                        : CommunicationObject(0, 0, 0),
                          m_timer(this, KTimer::High, KTimer::Foreground)
                      { // start it forever to run at 1 millisecond
                        // Do not forget to change the Communicationobject
                        // constructor when changing the timeout value to
                        // something else.
                        CommunicationObject::m_tickcondition = a_condition;
                        m_timer.start(1000);
                      }
    virtual           ~TimerCOB()
                      {
                      }

    // Overloaded virtuals from base class KTimerClient
    virtual void      timeOut(KTimer * /*a_timer*/)
                      { // Don't waist time on finding out which timer it was, we have only one
                        CommunicationObject::m_tickcount++;
                        m_tickcondition->signal();
                      }

    // Prevent this class from being abstract. We must be able to instantiate it
    virtual bool      create(void)
                      {
                        return true;
                      }
    virtual void      destroy(void)
                      {
                      }

  private:
    KTimer            m_timer;
};

// And now instantiate this class so the tickcount can start running
KCondition theOneAndOnlyTickcondition CYG_INIT_PRIORITY(APPLICATION);
TimerCOB theOneAndOnlyTimerCOB(&theOneAndOnlyTickcondition) CYG_INIT_PRIORITY(APPLICATION);

//----------------------
// Class implementations
//----------------------

// Class CommunicationObject
// Constructor
CommunicationObject::CommunicationObject(CAN_msgid a_cobid,
                                         can_io_handle_t a_handle,
                                         CommunicationObjectClient * a_client,
                                         unsigned int a_inhibittime)
    : m_client(a_client),
      m_cobid(a_cobid),
      m_marker(CommunicationObject::m_tickcount)
{
  _ASSERT(a_client != 0, "Invalid pointer passed");

  DWORD len;
  len = sizeof(CanMessageDispatcher *);
  can_io_get_config(a_handle, CAL_IO_GET_CONFIG_MSGDISPATCHER, &m_dispatcher, &len);
  len = sizeof(can_io_handle_t);
  can_io_get_config(a_handle, CAL_IO_GET_CONFIG_CANHANDLE, &m_handle, &len);

  // The following is also OK for a inhibittime of 0, ticks will also be set to 0
  // Do not forget to modify this when the timer value is changed from 1 millisecond
  if((a_inhibittime % 10) == 0)
    m_inhibitticks = (a_inhibittime / 10);
  else
    m_inhibitticks = (a_inhibittime / 10) + 1;
}

// Provide an (empty) virtual destructor
CommunicationObject::~CommunicationObject()
{
  m_dispatcher->detach(this);
}

// dataIndication
void CommunicationObject::dataIndication(unsigned char * a_data, unsigned char a_length)
{
  _ASSERT(a_length < 9,          "COB length outside valid range 0..8");
  _ASSERT(m_class != tx,         "Transmit COB received data");
  _ASSERT(m_class != RTR_server, "Transmit COB received data");

  m_client->dataIndication(a_data, a_length, this);
}

// waitForInhibit
void CommunicationObject::waitForInhibit(void)
{
  if(m_inhibitticks)
  { // only do this if inhibittime specified
    KSingleLock sl(m_tickcondition);
    while((m_marker + m_inhibitticks) > m_tickcount)
      sl.lock();

    m_marker = m_tickcount;
  }
}

// Class TxCOB
// Constructor
TxCOB::TxCOB(CAN_msgid a_cobid,
             can_io_handle_t a_handle,
             CommunicationObjectClient * a_client,
             unsigned int a_inhibittime)
    : CommunicationObject(a_cobid, a_handle, a_client, a_inhibittime)
{
  m_class = tx;
}

// Provide an (empty) virtual destructor
TxCOB::~TxCOB()
{
}

// create
bool TxCOB::create(void)
{ // We do not have to register the COB_id with the CAN driver. 
  // This should always be a success
  return true;
}

// destroy
void TxCOB::destroy(void)
{ // Nothing to do here
}

// transmit
bool TxCOB::transmit(const unsigned char * a_data, unsigned char a_length)
{
  _ASSERT(a_length < 9, "COB length outside valid range 0..8");

  can_message msg;
  msg.m_CAN_id = m_cobid;
  msg.m_RTR    = 0;
  msg.m_DLC    = a_length;
  memcpy(&msg.m_Data[0], a_data, a_length);

  DWORD len = sizeof(can_message);
  waitForInhibit();
  return(can_io_write(m_handle, &msg, &len) == ENOERR);
}

// Class RxCOB
// Constructor
RxCOB::RxCOB(CAN_msgid a_cobid,
             can_io_handle_t a_handle,
             CommunicationObjectClient * a_client)
    : CommunicationObject(a_cobid, a_handle, a_client)
{
  m_class = rx;
}

// Provide an (empty) virtual destructor
RxCOB::~RxCOB()
{
}

// create
bool RxCOB::create(void)
{
  m_dispatcher->attach(this);

  DWORD len = sizeof(CAN_msgid);
  return (can_io_set_config(m_handle, CAN_IO_ADD_RX_ID, &m_cobid, &len) == ENOERR);
}

// destroy
void RxCOB::destroy(void)
{
  DWORD len = sizeof(CAN_msgid);
  can_io_set_config(m_handle, CAN_IO_REMOVE_RX_ID, &m_cobid, &len);

  m_dispatcher->detach(this);
}

// Class RTR_TxCOB
// Constructor
RTR_ClientCOB::RTR_ClientCOB(CAN_msgid a_cobid,
                             can_io_handle_t a_handle,
                             CommunicationObjectClient * a_client,
                             unsigned int a_inhibittime)
    : CommunicationObject(a_cobid, a_handle, a_client, a_inhibittime)
{
  m_class = RTR_client;
}

// Provide an (empty) virtual destructor
RTR_ClientCOB::~RTR_ClientCOB()
{
}

// create
bool RTR_ClientCOB::create(void)
{
  m_dispatcher->attach(this);

  DWORD len = sizeof(CAN_msgid);
  return (can_io_set_config(m_handle, CAN_IO_ADD_RX_ID, &m_cobid, &len) == ENOERR);
}

// destroy
void RTR_ClientCOB::destroy(void)
{
  DWORD len = sizeof(CAN_msgid);
  can_io_set_config(m_handle, CAN_IO_REMOVE_RX_ID, &m_cobid, &len);

  m_dispatcher->detach(this);
}

// request
bool RTR_ClientCOB::request(unsigned char a_length)
{
  _ASSERT(a_length < 9, "COB length outside valid range 0..8");

  can_message msg;
  msg.m_CAN_id = m_cobid;
  msg.m_RTR    = 1;
  msg.m_DLC    = a_length;

  DWORD len = sizeof(can_message);
  waitForInhibit();
  return(can_io_write(m_handle, &msg, &len) == ENOERR);
}

// Class RTR_RxCOB
// Constructor
RTR_ServerCOB::RTR_ServerCOB(CAN_msgid a_cobid,
                             can_io_handle_t a_handle,
                             CommunicationObjectClient * a_client,
                             unsigned int a_inhibittime)
    : CommunicationObject(a_cobid, a_handle, a_client, a_inhibittime)
{
  m_class = RTR_server;
}

// Provide an (empty) virtual destructor
RTR_ServerCOB::~RTR_ServerCOB()
{
}

// create
bool RTR_ServerCOB::create(void)
{
  DWORD len = sizeof(CAN_msgid);
  return (can_io_set_config(m_handle, CAN_IO_ADD_RTR_ID, &m_cobid, &len) == ENOERR);
}

// destroy
void RTR_ServerCOB::destroy(void)
{
  DWORD len = sizeof(CAN_msgid);
  can_io_set_config(m_handle, CAN_IO_REMOVE_RTR_ID, &m_cobid, &len);
}

// update
bool RTR_ServerCOB::update(const unsigned char * a_data, unsigned char a_length)
{
  _ASSERT(a_length < 9, "COB length outside valid range 0..8");

  can_message msg;
  msg.m_CAN_id = m_cobid;
  msg.m_RTR = 0;
  msg.m_DLC = a_length;

  DWORD len = sizeof(can_message);
  memcpy(&msg.m_Data[0], a_data, a_length);

  return (can_io_set_config(m_handle, CAN_IO_UPDATE_RTR_BUFFER, &msg, &len) == ENOERR);
}

// transmit
bool RTR_ServerCOB::transmit(const unsigned char * a_data, unsigned char a_length)
{
  _ASSERT(a_length < 9, "COB length outside valid range 0..8");

  can_message msg;
  msg.m_CAN_id = m_cobid;
  msg.m_RTR    = 0;
  msg.m_DLC    = a_length;
  memcpy(&msg.m_Data[0], a_data, a_length);

  DWORD len = sizeof(can_message);

  if(can_io_set_config(m_handle, CAN_IO_UPDATE_RTR_BUFFER, &msg, &len) == ENOERR)
  {
    waitForInhibit();
    return(can_io_write(m_handle, &msg, &len) == ENOERR);
  }
  else
    return false;
}

#endif // COMOBJ_CXX
